@@ -8,7 +8,9 @@ package Net::Inspect::L7::HTTP;
 use base 'Net::Inspect::Flow';
 use Net::Inspect::Debug qw(:DEFAULT $DEBUG %TRACE);
 use Hash::Util 'lock_keys';
+use Digest::SHA 'sha1_base64';
 use Carp 'croak';
+use Scalar::Util 'weaken';
 use fields (
     'replay',   # collected and replayed in guess_protocol
     'meta',     # meta data from connection
@@ -36,7 +38,7 @@ our (@EXPORT_OK,%EXPORT_TAGS);
 	)]
     );
     push @EXPORT_OK,@$_ for (values %EXPORT_TAGS);
-    push @EXPORT_OK,'parse_hdrfields';
+    push @EXPORT_OK,'parse_hdrfields','parse_reqhdr','parse_rsphdr';
 }
 
 use constant {
@@ -54,6 +56,11 @@ use constant {
     RPBDY_DONE_ON_EOF => 0b10000,
 };
 
+my %Upgrade2Sub = (
+    'websocket' => \&upgrade_websocket,
+    'CONNECT'   => \&upgrade_CONNECT,
+);
+
 # rfc2616, 2.2
 #  token          = 1*<any CHAR except CTLs or separators>
 #  separators     = "(" | ")" | "<" | ">" | "@"
@@ -106,7 +113,7 @@ sub guess_protocol {
 	    return ($obj,$n);
 
 	} elsif ( $buf =~m{[^\n]\r?\n\r?\n}
-	    or length($buf)>$self->{hdr_maxsz}[0] ) {
+	    or length($buf)>2**16 ) {
 	    # does not look like a HTTP header for me
 	    debug("does not look like HTTP header: $buf");
 	    $guess->detach($self);
@@ -126,6 +133,7 @@ sub guess_protocol {
 
 {
     my $connid = 0;
+    sub syn { 1 }; # in case it is attached to Net::Inspect::Tcp
     sub new_connection {
 	my ($self,$meta,%args) = @_;
 	my $obj = $self->new;
@@ -209,7 +217,8 @@ sub _in0 {
 	croak 'existing error in connection' if $self->{error};
 
 	my $rqs = $self->{requests};
-	croak 'no open request' if ! @$rqs or $rqs->[0]{state} & RQBDY_DONE;
+	croak 'no open request' if ! @$rqs or
+	    $rqs->[0]{state} & RQBDY_DONE && ! $self->{upgrade};
 	croak 'existing error in request' if $rqs->[0]{state} & RQ_ERROR;
 	croak "gap too large" if $self->{gap_upto}[0]>=0
 	    && $self->{gap_upto}[0] < $self->{offset}[0] + $len;
@@ -222,16 +231,15 @@ sub _in0 {
 	}
 
 	$self->{offset}[0] += $len;
-	if ( my $obj = $rqs->[0]{obj} ) {
-	    if ($self->{upgrade}) {
-		$obj->in_data(0,[ gap => $len ],$eof,$time);
-	    } else {
-		$obj->in_request_body(
-		    [ gap => $len ],
-		    $eof || ($rqs->[0]{state} & RQBDY_DONE ? 1:0),
-		    $time
-		);
-	    }
+	my $obj = $rqs->[0]{obj};
+	if ($self->{upgrade}) {
+	    $self->{upgrade}[0]([ gap => $len ],$eof,$time);
+	} elsif ($obj) {
+	    $obj->in_request_body(
+		[ gap => $len ],
+		$eof || ($rqs->[0]{state} & RQBDY_DONE ? 1:0),
+		$time
+	    );
 	}
 	return $len;
     }
@@ -245,9 +253,7 @@ sub _in0 {
 
     if ($self->{upgrade}) {
 	$self->{offset}[0] += length($data);
-	if ( my $obj = $rqs->[0]{obj} ) {
-	    $obj->in_data(0,$data,$eof,$time);
-	}
+	$self->{upgrade}[0]($data,$eof,$time);
 	return $bytes + length($data);
     }
 
@@ -310,6 +316,8 @@ sub _in0 {
 	    rqchunked => undef,  # chunked mode for request
 	    rpchunked => undef,  # chunked mode for response
 	    method   => undef,   # request method
+	    upgrade  => undef,   # { websocket => sec-websocket-key }
+	    expect   => undef,   # hash with expectations like "100-continue"
 	    info     => '',      # debug info
 	);
 	lock_keys(%rq);
@@ -329,121 +337,48 @@ sub _in0 {
 	}
 
 	$DEBUG && $self->xdebug("need to read request header");
-	if ( $data =~s{( \A
-	    ([A-Z]{2,20})[\040\t]+          # method
-	    (\S+)[\040\t]+                  # path/URI
-	    HTTP/(1\.[01])[\40\t]*          # version
-	    \r?\n                           # (CR)LF
-	    ([^\r\n].*?\n)?                 # fields
-	    \r?\n                           # final (CR)LF
-	)}{}sxi ) {
+	if ($data =~s{\A(\A.*?\n\r?\n)}{}s) {
+	    $DEBUG && $self->xdebug("got request header");
 	    my $hdr = $1;
-	    $rq->{method} = uc($2);
-	    my $url = $3;
-	    my $version = $4;
-	    $rq->{info} = "\U$2\E $3 HTTP/$4";
-	    $DEBUG && $self->xdebug("got request header $rq->{info}");
-
-	    my %kv;
-	    my $bad = parse_hdrfields($5,\%kv);
-	    %TRACE && ($obj||$self)->xtrace("invalid request header data: $bad") 
-		if $bad ne '';
-
 	    my $n = length($hdr);
 	    $self->{offset}[0] += $n;
 	    $bytes += $n;
 	    $rq->{state} |= RQHDR_DONE; # rqhdr done
 
-	    if ( my $cl = $kv{'content-length'} ) {
-		if ( @$cl>1 and do { my %x; @x{@$cl} = (); keys(%x) } > 1 ) {
-		    ($obj||$self)->fatal(
-			"multiple different content-length header in request",
-			0,$time);
-		    $rq->{state} |= RQ_ERROR;
-		    return $bytes;
-		}
-		if ($cl->[0] !~m{^(\d+)$}) {
-		    ($obj||$self)->fatal(
-			"invalid content-length '$cl->[0]' in request",
-			0,$time);
-		    $rq->{state} |= RQ_ERROR;
-		    return $bytes;
-		}
-		$rq->{rqclen} = $cl->[0];
-		$DEBUG && $self->xdebug(
-		    "set content-length to $rq->{rqclen} from header");
+	    my %hdr;
+	    my $err = parse_reqhdr($hdr,\%hdr);
+	    if ($err) {
+		($obj||$self)->fatal($err,0,$time);
+		$rq->{state} |= RQ_ERROR;
+		return $bytes;
 	    }
 
-	    if ( $version >= 1.1 and 
-		grep { m{(?:^|[ \t,])chunked(?:$|[ \t,;])}i }
-		    @{ $kv{'transfer-encoding'} || [] }
-	    ) {
+	    my $body_done;
+	    if ($hdr{chunked}) {
 		$rq->{rqchunked} = 1;
-		if ( defined $rq->{rqclen} ) {
-		    # RFC2616 4.4.3: if both given ignore content-length
-		    %TRACE && ($obj||$self)->xtrace(
-			"request content-length and chunked given");
-		    $rq->{rqclen} = undef;
-		}
+	    } elsif ($hdr{content_length}) {
+		$rq->{rqclen} = $hdr{content_length};
+		$self->{gap_upto}[0]= $self->{offset}[0] + $hdr{content_length};
+	    } else {
+		$body_done = $hdr{method} ne 'CONNECT';
 	    }
 
-	    if ( $METHODS_WITHOUT_RQBODY{ $rq->{method} } ) {
-		if ( $rq->{rqclen} ) {
-		    ($obj||$self)->fatal(
-			"no body allowed with method $rq->{method}",0,$time);
-		    $rq->{state} |= RQ_ERROR;
-		    return $bytes;
-		}
-		# With CONNECT no data are allowed before we got the successful
-		# response. Only then rqclen will be set to undef to mark that
-		# body ends with eof.
-		$rq->{rqclen} = 0;
-		$rq->{rqchunked} = undef;
-	    } elsif ( $METHODS_WITH_RQBODY{ $rq->{method} } ) {
-		if ( ! defined $rq->{rqclen} && ! $rq->{rqchunked} ) {
-		    ($obj||$self)->fatal(
-			"content-length or transfer-encoding chunked must be given with method $rq->{method}",
-			0,$time);
-		    $rq->{state} |= RQ_ERROR;
-		    return $bytes;
-		}
-	    } elsif ( ! $rq->{rqchunked} ) {
-		# if not given content-length is considered 0
-		$rq->{rqclen} ||= 0;
-	    }
+	    $rq->{expect}  = $hdr{expect};
+	    $rq->{method}  = $hdr{method};
+	    $rq->{info}    = $hdr{info};
+	    $rq->{upgrade} = $hdr{upgrade};
 
-	    $obj && $obj->in_request_header($hdr,$time, { 
-		content_length => $rq->{rqclen},
-		$rq->{rqchunked} ? ( chunked => 1 ):(),
-		method => $rq->{method},
-		url => $url,
-		version => $version,
-		fields => \%kv,
-		$bad eq '' ? () : ( junk => $bad ),
-	    });
+	    %TRACE && $hdr{junk} && ($obj||$self)->xtrace(
+		"invalid request header data: $hdr{junk}");
 
-	    if (!$rq->{rqchunked}) {
-		if (!defined($rq->{rqclen})) {
-		    $self->{gap_upto}[0] = -1; # until eof (CONNECT)
-		} elsif ($rq->{rqclen}) {
-		    $self->{gap_upto}[0] = $self->{offset}[0] + $rq->{rqclen};
-		}
-	    }
+	    $obj && $obj->in_request_header($hdr,$time,\%hdr);
 
-	    if ( ! $rq->{rqchunked}
-		and $rq->{rqclen} == 0
-		and $rq->{method} ne 'CONNECT') {
+	    if ($body_done) {
 		$DEBUG && $self->xdebug("no content-length - request body done");
-		# no clen - body done
 		$rq->{state} |= RQBDY_DONE;
 		$obj && $obj->in_request_body('',1,$time);
 	    }
 
-	} elsif ( $data =~m{[^\n]\r?\n\r?\n}g ) {
-	    ($obj||$self)->fatal( sprintf("invalid request header syntax '%s'",
-		substr($data,0,pos($data))),0,$time);
-	    $rq->{state} |= RQ_ERROR;
-	    return $bytes;
 	} elsif ( length($data) > $self->{hdr_maxsz}[0] ) {
 	    ($obj||$self)->fatal('request header too big',0,$time);
 	    $rq->{state} |= RQ_ERROR;
@@ -602,18 +537,17 @@ sub _in1 {
 	$rq->{rpclen} -= $len if defined $rq->{rpclen};
 	$self->{offset}[1] += $len;
 
-	if ( my $obj = $rq->{obj} ) {
-	    if ($self->{upgrade}) {
-		$obj->in_data(1,[ gap => $len ],$eof,$time);
-	    } elsif ($rq->{rpclen}
-		or !defined $rq->{rpclen}
-		or $rq->{rpchunked}) {
-		$obj->in_response_body([ gap => $len ],$eof,$time);
-	    } else {
-		# done with request
-		pop(@$rqs);
-		$obj->in_response_body([ gap => $len ],1,$time);
-	    }
+	my $obj = $rq->{obj};
+	if ($self->{upgrade}) {
+	    $self->{upgrade}[1]([ gap => $len ],$eof,$time);
+	} elsif ($rq->{rpclen}
+	    or !defined $rq->{rpclen}
+	    or $rq->{rpchunked}) {
+	    $obj && $obj->in_response_body([ gap => $len ],$eof,$time);
+	} else {
+	    # done with request
+	    pop(@$rqs);
+	    $obj && $obj->in_response_body([ gap => $len ],1,$time);
 	}
 	return $len;
     }
@@ -622,19 +556,16 @@ sub _in1 {
     READ_DATA:
 
     return $bytes if $self->{error};
+    return $bytes if $data eq '' && !$eof;
 
     if ($self->{upgrade}) {
-	return $bytes if $data eq '' && !$eof;
 	$self->{offset}[1] += length($data);
-	if ( my $obj = $rqs->[0]{obj} ) {
-	    $obj->in_data(1,$data,$eof,$time);
-	}
+	$self->{upgrade}[1]($data,$eof,$time);
 	return $bytes + length($data);
     }
 
     if ( $data eq '' ) {
 	$DEBUG && $self->xdebug("no more data, eof=$eof bytes=$bytes");
-	return $bytes if ! $eof; # need more data
 
 	# handle EOF
 	# check if we got response body for last request
@@ -681,109 +612,78 @@ sub _in1 {
 	}
 
 	# no response header yet, check if data contains it
-	if ( $data =~s{( \A
-	    HTTP/(1\.[01])[\040\t]+          # version
-	    (\d\d\d)                         # code
-	    (?:[\040\t]+([^\r\n]*))?         # reason
-	    \r?\n
-	    ([^\r\n].*?\n)?                  # fields
-	    \r?\n                            # empty line
-	)}{}sxi) {
-	    my ($hdr,$version,$code,$reason) = ($1,$2,$3,$4);
-	    my %kv;
-	    my $bad = parse_hdrfields($5,\%kv);
-	    %TRACE && ($obj||$self)->xtrace("invalid response header data: $bad") 
-		if $bad ne '';
-
+	if ( $data =~s{\A(.*?\n\r?\n)}{}s ) {
+	    my $hdr = $1;
 	    my $n = length($hdr);
 	    $bytes += $n;
 	    $self->{offset}[1] += $n;
 
-	    if ( $code >= 100 and $code <= 199 ) {
-		# preliminary response, we are not done!
-		$DEBUG && $self->xdebug("got preliminary response");
-		$obj && $obj->in_response_header($hdr,$time,{
-		    content_length => 0,
-		    version => $version,
-		    code => $code,
-		    reason => $reason,
-		    fields => \%kv,
-		    $bad eq '' ? () : ( junk => $bad ),
-		});
+	    my %hdr;
+	    my $err = parse_rsphdr($hdr, {
+		method  => $rq->{method},
+		upgrade => $rq->{upgrade},
+		expect  => $rq->{expect},
+	    }, \%hdr);
 
-		# preliminary responses do not contain any body
+	    goto error if $err;
+	    $DEBUG && $self->xdebug("got response header");
+
+	    %TRACE && $hdr{junk} && ($obj||$self)->xtrace(
+		"invalid request header data: $hdr{junk}");
+
+	    if ($hdr{preliminary}) {
+		# Preliminary response. Wait for read real response.
+		$obj && $obj->in_response_header($hdr,$time,\%hdr);
 		goto READ_DATA;
 	    }
 
 	    $rq->{state} |= RPHDR_DONE; # response header done
 
+	    if ($hdr{upgrade}) {
+		# Reset length to undef since we need to read until eof.
+		$rq->{rpclen} = undef;
+		$self->{upgrade} = $Upgrade2Sub{$hdr{upgrade}}($self,$obj)
+		or do {
+		    $err = "invalid or unsupported connection upgrade";
+		    goto error;
+		};
+		goto done;
+	    }
 
-	    if ( $rq->{method} eq 'CONNECT' and $code =~m{^2} ) {
-		$self->{upgrade} = 1;
-		# Since CONNECT was successful we can now allow unlimited
-		# request and response body data.
-		$rq->{rqclen} = $rq->{rpclen} = undef;
-		@{$self->{gap_upto}} = (-1,-1);
+	    # Forget any $rq->{upgrade} since we did not get 101 response.
+	    # This might be due to authorization required etc.
+	    $rq->{upgrade} = undef;
 
-	    } elsif ( $METHODS_WITHOUT_RPBODY{ $rq->{method} }
-		or $CODE_WITHOUT_RPBODY{$code} ) {
-		# no content, even if specified
-		$rq->{rpclen} = 0;
-
-	    } elsif ( $version >= 1.1 and
-		grep { m{(?:^|[ \t,])chunked(?:$|[ \t,;])}i }
-		    @{ $kv{'transfer-encoding'} || [] }
-	    ) {
-		# RFC2616 4.4.3:
-		# chunked takes preference if content-length is given too.
+	    my $body_done;
+	    if ($hdr{chunked}) {
 		$rq->{rpchunked} = 1;
-
-	    } elsif ( my $cl = $kv{'content-length'} ) {
-		if ( @$cl>1 and do { my %x; @x{@$cl} = (); keys(%x) } > 1 ) {
-		    ($obj||$self)->fatal(
-			"multiple different content-length header in response",
-			1,$time);
-		    $self->{error} = 1;
-		    return $bytes;
+	    } elsif (defined $hdr{content_length}) {
+		if (($rq->{rpclen} = $hdr{content_length})) {
+		    # content_length > 0, can do gaps
+		    $self->{gap_upto}[1]= $self->{offset}[1]
+			+ $hdr{content_length};
+		} else {
+		    $body_done = 1;
 		}
-		if ($cl->[0] !~m{^(\d+)$}) {
-		    ($obj||$self)->fatal(
-			"invalid content-length '$cl->[0]' in response",
-			1,$time);
-		    $self->{error} = 1;
-		    return $bytes;
-		}
-		$rq->{rpclen} = $cl->[0];
-		$self->{gap_upto}[1] = $self->{offset}[1] + $rq->{rpclen};
-
 	    } else {
 		# no length given but method supports body -> end with eof
+		$rq->{state} |= RPBDY_DONE_ON_EOF; # body done when eof
 		$self->{gap_upto}[1] = -1;
 	    }
 
-	    $DEBUG && $self->xdebug("got response header");
-	    $obj && $obj->in_response_header($hdr,$time,{
-		content_length => $rq->{rpclen},
-		$rq->{rpchunked} ? ( chunked => 1 ):(),
-		version => $version,
-		code => $code,
-		reason => $reason,
-		fields => \%kv,
-		$bad eq '' ? () : ( junk => $bad ),
-	    });
-
-	    # if no body invoke hook with empty body and eof
-	    if ( defined $rq->{rpclen} and $rq->{rpclen} == 0 ) {
-		# clen == 0 -> body done
+	    done:
+	    $obj && $obj->in_response_header($hdr,$time,\%hdr);
+	    if ($body_done) {
 		$DEBUG && $self->xdebug("no response body");
 		pop(@$rqs);
 		$obj && $obj->in_response_body('',1,$time);
-		goto READ_DATA;
 	    }
+	    goto READ_DATA;
 
-	    if ( ! $rq->{rpchunked} && ! $rq->{rpclen} ) {
-		$rq->{state} |= RPBDY_DONE_ON_EOF; # body done when eof
-	    }
+	    error:
+	    $self->{error} = 1;
+	    ($obj||$self)->fatal($err,1,$time);
+	    return $bytes;
 
 	} elsif ( $data =~m{[^\n]\r?\n\r?\n}g ) {
 	    ($obj||$self)->fatal( sprintf("invalid response header syntax '%s'",
@@ -961,6 +861,210 @@ sub parse_hdrfields {
     return $bad;
 }
 
+sub parse_reqhdr {
+    my ($data,$hdr) = @_;
+    $data =~m{\A
+	([A-Z]{2,20})[\040\t]+          # $1: method
+	(\S+)[\040\t]+                  # $2: path/URI
+	HTTP/(1\.[01])[\40\t]*          # $3: version
+	\r?\n                           # (CR)LF
+	([^\r\n].*?\n)?                 # $4: fields
+	\r?\n                           # final (CR)LF
+    \Z}sx or return "invalid request header";
+
+    my $version = $3;
+    my $method  = $1;
+    %$hdr = (
+	method    => $method,
+	url       => $2,
+	version   => $version,
+	info      => "$method $2 HTTP/$version",
+	# fields  -  hash of fields
+	# junk    -  bad header fields
+	# expect  -  expectations from expect header
+	# upgrade -  { websocket => key }
+	# content_length
+	# chunked
+    );
+
+    my %kv;
+    my $bad = parse_hdrfields($4,\%kv);
+    $hdr->{junk} = $bad if $bad ne '';
+    $hdr->{fields} = \%kv;
+
+    if ($version>=1.1 and $kv{expect}) {
+	for(@{$kv{expect}}) {
+	    # ignore all but 100-continue
+	    $hdr->{expect}{lc($1)} = 1 if m{\b(100-continue)\b}i
+	}
+    }
+
+    # RFC2616 4.4.3:
+    # chunked transfer-encoding takes preferece before content-length
+    if ( $version >= 1.1 and
+	grep { m{(?:^|[ \t,])chunked(?:$|[ \t,;])}i }
+	    @{ $kv{'transfer-encoding'} || [] }
+    ) {
+	$hdr->{chunked} = 1;
+
+    } elsif ( my $cl = $kv{'content-length'} ) {
+	return "multiple different content-length header in request"
+	    if @$cl>1 and do { my %x; @x{@$cl} = (); keys(%x) } > 1;
+	return "invalid content-length '$cl->[0]' in request"
+	    if $cl->[0] !~m{^(\d+)$};
+	$hdr->{content_length} = $cl->[0];
+    }
+
+    if ( $METHODS_WITHOUT_RQBODY{$method} ) {
+	# Ignore any kind of length information.
+	$hdr->{content_length} = 0;
+	delete $hdr->{chunked};
+
+    } elsif ( $METHODS_WITH_RQBODY{$method} ) {
+	return "content-length or transfer-encoding chunked must be given with method $method"
+	    if ! $hdr->{chunked} and ! defined $hdr->{content_length};
+
+    } elsif ( ! $hdr->{chunked} ) {
+	# if not given content-length is considered 0
+	$hdr->{content_length} ||= 0;
+    }
+
+    # Connection upgrade - currently only Websocket is supported
+    if ($version >= 1.1 and $kv{upgrade}
+	and grep m{\bWebSocket\b}i,@{$kv{upgrade}}) {
+	# Websocket: RFC6455, Sec.4.1 Page 16ff
+	my $wskey = $kv{'sec-websocket-key'} || [];
+	if (@$wskey > 1) {
+	    my %x;
+	    $wskey = [ map { $x{$_}++ ? ():($_) } @$wskey ];
+	}
+	my $v;
+	if ( @$wskey == 1
+	    and $method eq 'GET'
+	    # RFC6455 requires Connection upgrade on client site,
+	    # while for RFC7230 it should be enough on server side.
+	    and $v = $kv{connection} and grep m{\bUPGRADE\b}i, @$v
+	    and $v = $kv{upgrade} and grep m{\bWebSocket\b}i, @$v
+	    and $v = $kv{'sec-websocket-version'}
+	    and ! grep { $_ ne '13' } @$v
+	) {
+	    $hdr->{upgrade} = { websocket => $wskey->[0] };
+	} else {
+	    return "invalid websocket upgrade request";
+	}
+    }
+    return; # no error
+}
+
+sub parse_rsphdr {
+    my ($data,$request,$hdr) = @_;
+    $data =~ m{\A
+	HTTP/(1\.[01])[\040\t]+          # $1: version
+	(\d\d\d)                         # $2: code
+	(?:[\040\t]+([^\r\n]*))?         # $3: reason
+	\r?\n
+	([^\r\n].*?\n)?                  # $4: fields
+	\r?\n                            # empty line
+    \Z}sx or return "invalid response header";
+
+    my $version = $1;
+    my $code = $2;
+    %$hdr = (
+	version   => $version,
+	code      => $code,
+	reason    => $3,
+	# fields
+	# junk
+	# content_length
+	# chunked
+	# upgrade
+	# preliminary
+    );
+
+    my %kv;
+    my $bad = parse_hdrfields($4,\%kv);
+    $hdr->{fields} = \%kv;
+    $hdr->{junk} = $bad if $bad ne '';
+
+    if ($code == 100 and $request->{expect}{'100-continue'}
+	or $code == 102) {
+	# Preliminary responses do not contain any body.
+	# 100 should only happen with Expect: 100-continue from client
+	$hdr->{preliminary} = 1;
+	$hdr->{content_length} = 0;
+    } elsif ($code != 101 and $code <= 199) {
+	return "unexpected status code $code";
+    }
+
+    # Switching Protocols
+    # Any upgrade must have both a "Connection: upgrade" and a
+    # "Upgrade: newprotocol" header.
+    if ($code == 101) {
+	my %proto;
+	if ($request->{upgrade}
+	    and grep { m{\bUPGRADE\b}i } @{$kv{connection} || []}) {
+	    for(@{$kv{upgrade} || []}) {
+		$proto{lc($_)} = 1 for split(m{\s*[,;]\s*});
+	    }
+	}
+
+	# Currently only Websocket is supported.
+	if (keys(%proto) == 1
+	    and $proto{websocket}
+	    and my $wsk = $request->{upgrade}{websocket}
+	    and my $wsa = $kv{'sec-websocket-accept'}) {
+	    if ($wsa && @$wsa != 1) {
+		my %x;
+		$wsa = [ map { $x{$_}++ ? ():($_) } @$wsa ];
+	    }
+	    # beware its magic! see RFC6455 page 7
+	    # sha1_base64 does no padding, so we need to add a single
+	    # '=' (pad to 4*7 byte) at the end for comparison
+	    if ( @$wsa != 1 or $wsa->[0] ne sha1_base64(
+		$wsk.'258EAFA5-E914-47DA-95CA-C5AB0DC85B11').'=') {
+		return "invalid websocket upgrade response";
+	    }
+	    $hdr->{upgrade} = 'websocket';
+	    return;
+
+	} else {
+	    return "invalid or unsupported connection upgrade";
+	}
+    }
+
+    # successful response to CONNECT
+    if ($request->{method} eq 'CONNECT' and $code >= 200 and $code < 300) {
+	$hdr->{upgrade} = 'CONNECT';
+	return;
+    }
+
+    # RFC2616 4.4.3:
+    # chunked transfer-encoding takes preferece before content-length
+    if ( $version >= 1.1 and
+	grep { m{(?:^|[ \t,])chunked(?:$|[ \t,;])}i }
+	    @{ $kv{'transfer-encoding'} || [] }
+    ) {
+	$hdr->{chunked} = 1;
+
+    } elsif ( my $cl = $kv{'content-length'} ) {
+	return "multiple different content-length header in response"
+	    if @$cl>1 and do { my %x; @x{@$cl} = (); keys(%x) } > 1;
+	return "invalid content-length '$cl->[0]' in response"
+	    if $cl->[0] !~m{^(\d+)$};
+	$hdr->{content_length} = $cl->[0];
+    }
+
+    if ($CODE_WITHOUT_RPBODY{$code}
+	or $METHODS_WITHOUT_RPBODY{$request->{method}}) {
+	# no content, even if specified
+	$hdr->{content_length} = 0;
+	delete $hdr->{chunked};
+	return;
+    }
+
+    return;
+}
+
 
 sub new_request {
     my $self = shift;
@@ -1013,6 +1117,313 @@ sub dump_state {
     $self->xdebug($msg);
 }
 
+sub upgrade_CONNECT {
+    my ($self,$obj) = @_;
+    my @sub;
+    if ($obj) {
+	weaken($obj);
+	for my $dir (0,1) {
+	    my $dir = $dir; # old $dir was alias only
+	    $sub[$dir] = sub {
+		my ($data,$eof,$time) = @_;
+		$obj->in_data($dir,$data,$eof,$time);
+		return ref($data) ? $data->[1] : length($data)
+	    }
+	};
+    } else {
+	$sub[0] = $sub[1] = sub {
+	    my $data = shift;
+	    return ref($data) ? $data->[1] : length($data)
+	};
+    }
+    @{$self->{gap_upto}} = (-1,-1);
+    return \@sub;
+}
+
+sub upgrade_websocket {
+    my ($self,$obj) = @_;
+    goto &upgrade_CONNECT if !$obj || !$obj->can('in_wsdata');
+
+    weaken($obj);
+    weaken($self);
+
+    my @sub;
+    for my $dir (0,1) {
+	my $dir = $dir; # old $dir is only alias
+	my $rbuf = '';
+
+	# If $clen is defined we are inside a frame ($current_frame).
+	# If $clen is not defined all other variables here do not matter.
+	# Since control messages might be in-between fragmented data messages we
+	# need to keep this information for an open data message.
+	my ($clen,$clenhi,$current_frame,$data_frame,$ctl_frame,$got_close);
+
+	$sub[$dir] = sub {
+	    my ($data,$eof,$time) = @_;
+	    my $err;
+
+	    # Handle data gaps. These are only allowed inside data frames.
+	    ############################################################
+	    if (ref($data)) {
+		croak "unknown type $data->[0]" if $data->[0] ne 'gap';
+		my $gap = $data->[1];
+		if (!defined $clen) {
+		    $err = "gap outside websocket frame";
+		    goto bad;
+		}
+		if (!$data_frame || $current_frame != $data_frame) {
+		    $err = "gap inside control frame";
+		    goto bad;
+		}
+		my $eom = 0; # end of message on end-of-frame + FIN frame
+		while ($gap>0) {
+		    if ($clen == 0) {
+			if (!$clenhi) {
+			    $err = "gap larger than frame size";
+			    goto bad;
+			}
+			$clenhi--;
+			$clen = 0xffffffff;
+			$gap--;
+			$current_frame->{mask_offset}
+			    = (($current_frame->{mask_offset}||0) + 1) % 4;
+
+		    } elsif ($gap > $clen) {
+			$gap -= $clen;
+			$current_frame->{mask_offset}
+			    = (($current_frame->{mask_offset}||0) + $clen) % 4;
+			$clen = 0;
+		    } else { # $gap <= $clen
+			$clen -= $gap;
+			$current_frame->{mask_offset}
+			    = (($current_frame->{mask_offset}||0) + $gap) % 4;
+			$gap = 0;
+		    }
+		}
+		if (!$clen && !$clenhi) {
+		    # frame done
+		    $eom = $data_frame->{fin} ? 1:0;
+		    $clen = undef;
+		}
+
+		$obj->in_wsdata($dir,$data,$eom,$time,$data_frame);
+		delete $data_frame->{init};
+		if ($eom) {
+		    $data_frame = $current_frame = undef;
+		    $self->{gap_upto}[$dir] = 0;
+		}
+		return;
+	    }
+
+	    $rbuf .= $data;
+
+	    PARSE_DATA:
+
+	    # data for existing frame
+	    ############################################################
+	    if (defined $clen) {
+		my $size = length($rbuf);
+		if (!$size and $clen || $clenhi) {
+		    goto done if ! $eof;
+		    $err = "eof inside websocket frame";
+		    goto bad;
+		}
+		my $fwd = '';
+		my $eom = 0;
+		while ($size>0) {
+		    if ($clen == 0) {
+			last if !$clenhi;
+			$clenhi--;
+			$clen = 0xffffffff;
+			$size--;
+			$fwd .= substr($rbuf,0,1,'');
+		    } elsif ($size > $clen) {
+			$size -= $clen;
+			$fwd .= substr($rbuf,0,$clen,'');
+			$clen = 0;
+		    } else {  # $size < $clen
+			$clen -= $size;
+			$size = 0;
+			$fwd .= $rbuf;
+			$rbuf = '';
+		    }
+		}
+		if (!$clen && !$clenhi) {
+		    # frame done
+		    $eom = $current_frame->{fin} ? 1:0;
+		    $clen = undef;
+		}
+		if ($data_frame && $current_frame == $data_frame) {
+		    $obj->in_wsdata($dir,$fwd,$eom,$time,$data_frame);
+		    if ($eom) {
+			$data_frame = undef;
+		    } else {
+			delete $data_frame->{init};
+			$current_frame->{mask_offset}
+			    = (($current_frame->{mask_offset}||0) + length($fwd)) % 4
+			    if defined $clen;
+		    }
+		} else {
+		    die "expected to read full control frame" if defined $clen;
+		    if ($current_frame->{opcode} == 0x8) {
+			# extract status + reason for close
+			if ($fwd eq '') {
+			    $current_frame->{status} = 1005; # RFC6455, 7.1.5
+			} elsif (length($fwd) < 2) {
+			    # if payload it must be at least 2 byte for status
+			    $err = "invalid length for close control frame";
+			    goto bad;
+			} else {
+			    ($current_frame->{status},$current_frame->{reason})
+				= unpack("na*",$current_frame->unmask($fwd));
+			}
+		    }
+		    $obj->in_wsctl($dir,$fwd,$time,$current_frame);
+		}
+		goto done if !$size;
+		goto PARSE_DATA;
+	    }
+
+	    # start of new frame: read frame header
+	    ############################################################
+	    goto hdr_need_more if length($rbuf)<2;
+
+	    (my $flags,$clen) = unpack("CC",$rbuf);
+	    my $mask = $clen & 0x80;
+	    $clen &= 0x7f;
+	    $clenhi = 0;
+	    my $off = 2;
+
+	    if ($clen == 126) {
+		goto hdr_need_more if length($rbuf)<4;
+		($clen) = unpack("xxn",$rbuf);
+		goto bad_length if $clen<126;
+		$off = 4;
+	    } elsif ($clen == 127) {
+		goto hdr_need_more if length($rbuf)<10;
+		($clenhi,$clen) = unpack("xxNN",$rbuf);
+		goto bad_length if !$clenhi && $clen<2**16;
+		$off = 10;
+	    }
+	    if ($mask) {
+		goto hdr_need_more if length($rbuf)<$off+4;
+		($mask) = unpack("x${off}a4",$rbuf);
+		$off+=4;
+	    } else {
+		$mask = undef;
+	    }
+
+	    my $opcode = $flags & 0b00001111;
+	    my $fin    = $flags & 0b10000000;
+	    goto reserved_flag if $flags & 0b01110000;
+
+	    if ($opcode >= 0x8) {
+		# control frame
+		goto reserved_opcode if $opcode >= 0xb;
+		if (!$fin) {
+		    $err = "fragmented control frames are forbidden";
+		    goto bad;
+		}
+		if ($clenhi || $clen>125) {
+		    $err = "control frames should be <= 125 bytes";
+		    goto bad;
+		}
+		# We like to forward control frames as a single entity, so make
+		# sure we get the whole (small) frame at once.
+		goto hdr_need_more if $off+$clen > length($rbuf);
+
+		$current_frame = $ctl_frame
+		    ||= Net::Inspect::L7::HTTP::_WSFrame->new;
+		%$current_frame = (
+		    opcode => $opcode,
+		    defined($mask) ? ( mask => $mask ):()
+		);
+		$got_close = 1 if $opcode == 0x8;
+
+	    } elsif ($opcode>0) {
+		# data frame, but no continuation
+		goto reserved_opcode if $opcode >= 0x3;
+		if ($got_close) {
+		    $err = "data frame after close";
+		    goto bad;
+		}
+		$current_frame = $data_frame
+		    ||= Net::Inspect::L7::HTTP::_WSFrame->new;
+		%$current_frame = (
+		    opcode => $opcode,
+		    $fin ? ( fin => 1 ):(),
+		    init => 1,  # initial data
+		    defined($mask) ? ( mask => $mask ):()
+		);
+
+	    } else {
+		# continuation frame
+		if (!$data_frame) {
+		    $err = "continuation frame without previous data frame";
+		    goto bad;
+		}
+		$current_frame = $data_frame;
+		%$current_frame = (
+		    opcode => $data_frame->{opcode},
+		    $fin ? ( fin => 1 ):(),
+		    defined($mask) ? ( mask => $mask ):()
+		);
+	    }
+
+	    # done with frame header
+	    substr($rbuf,0,$off,'');
+	    goto PARSE_DATA;
+
+	    # Done
+	    ############################################################
+
+	    hdr_need_more:
+	    $clen = undef; # re-read from start if frame next time
+
+	    done:
+	    if (defined $clen) {
+		# Processed all inside rbuf we want more (data frame)
+		$self->{gap_upto}[$dir] = $self->{offset}[$dir] + (
+		    ! $clenhi ? $clen :           # len <=32 bit
+		    1 << 32 == 1 ? 0xffffffff :   # platform 32bit only
+		    ($clenhi << 32) + $clen       # full 64 bit
+		);
+	    }
+	    if ($eof) {
+		# forward eof as special wsctl with no frame
+		$obj->in_wsctl($dir,'',$time);
+	    }
+	    return;
+
+	    bad_length:
+	    $err ||= "non-minimal length representation in websocket frame";
+	    reserved_flag:
+	    $err ||= "extensions using reserved flags are not supported";
+	    reserved_opcode:
+	    $err ||= "no support for opcode $opcode";
+
+	    bad:
+	    $self->{error} = 1;
+	    $obj->fatal($err,$dir,$time);
+	    return;
+	};
+    }
+    return \@sub;
+}
+
+{
+    package Net::Inspect::L7::HTTP::_WSFrame;
+    sub new { bless {}, shift };
+    sub unmask {
+	my ($self,$data) = @_;
+	return $data if $data eq '' or ! $self->{mask};
+	my $l = length($data);
+	$data ^= substr($self->{mask} x int($l/4+2),$self->{mask_offset}||0,$l);
+	return $data;
+    };
+}
+
+
 1;
 
 __END__
@@ -1102,6 +1513,8 @@ $header is the string of the header.
 
 =item version - version of HTTP spoken in request
 
+=item info - first line of request (method url version)
+
 =item fields - (key => \@values) hash of header fields
 
 =item junk - invalid data found in header fields part
@@ -1110,6 +1523,15 @@ $header is the string of the header.
 
 =item chunked - true if body uses transfer encoding chunked
 
+=item upgrade - contains hash when protocol upgrade was requested
+
+Currently this hash contains the key C<websocket> with the value of the
+C<sec-websocket-key> if a valid request for a Websocket upgrade was detected.
+
+=item expect - contains hash for expectations from Expect header
+
+Currently the only possible key is C<100-continue>.
+
 =back
 
 
@@ -1136,6 +1558,10 @@ $header is the string of the header.
 
 =item chunked - true if body uses transfer encoding chunked
 
+=item upgrade - new protocol when switching protocols, e.g. 'websocket'
+
+=item preliminary - true if this is a preliminary response
+
 =back
 
 =item $request->in_request_body($data,$eobody,$time)
@@ -1143,8 +1569,8 @@ $header is the string of the header.
 Called for a chunk of data of the request body.
 $eobody is true if this is the last chunk of the request body.
 If the request body is empty the method will be called once with C<''>.
-If no body exists because of CONNECT or HTTP Upgrade C<in_data> will be called,
-not C<in_request_body>.
+If no body exists because of CONNECT or HTTP Upgrade C<in_data> or the websocket
+functions will be called, not C<in_request_body>.
 
 $data can be C<< [ 'gap' => $len ] >> if the input to this
 layer were gaps.
@@ -1155,8 +1581,8 @@ Called for a chunk of data of the response body.
 $eof is true if this is the last chunk of the connection.
 $eobody is true if this is the last chunk of the response body.
 If the response body is empty the method will be called once with C<''>.
-If no body exists because of CONNECT or HTTP Upgrade C<in_data> will be called,
-not C<in_response_body>.
+If no body exists because of CONNECT or HTTP Upgrade C<in_data> or the websocket
+functions will be called, not C<in_response_body>.
 
 $data can be C<< [ 'gap' => $len ] >> if the input to this
 layer were gaps.
@@ -1177,9 +1603,42 @@ Will be called after in_response_body/in_request_body got called with eof true.
 
 =item $request->in_data($dir,$data,$eof,$time)
 
-Will be called for any data after successful CONNECT or Upgrade, Websockets...
+Will be called for any data after successful CONNECT or Upgrade.
+If no websocket functions are defined in the request object it will also be used
+for Websockets.
 C<$dir> is 0 for data from client, 1 for data from server.
 
+=item $request->in_wsctl($dir,$data,$time,$frameinfo)
+
+This will be called after a Websocket upgrade when receiving a control frame.
+C<$dir> is 0 for data from client, 1 for data from server.
+C<$data> is the unmasked payload of the frame.
+C<$frameinfo> is a blessed hash reference which contains the C<opcode> of the
+frame and the C<mask> (binary). For a close frame it will also contain the
+extracted C<status> code and the C<reason>.
+
+To get the unmasked payload call C<< $frameinfo->unmask($masked_data) >>.
+
+C<in_wsctl> will be called on connection close with C<$data> of C<''> and no
+C<\%frameinfo> (i.e. no hash reference).
+
+=item $request->in_wsdata($dir,$data,$eom,$time,$frameinfo)
+
+This will be called after a Websocket upgrade when receiving data inside a data
+frame. Contrary to (the short) control frames the data frame must not be read
+fully before calling C<in_wsdata>.
+
+C<$dir> is 0 for data from client, 1 for data from server.
+C<$data> is the unmasked payload of the frame.
+C<$eom> is true if the message is done with this call, that is if the data frame
+is done and the FIN bit was set on the frame.
+C<$frameinfo> is a blessed hash reference which contains the data type as
+C<opcode>. This will be the original opcode of the starting frame in case of
+fragmented transfer. It will also contain the C<mask> (binary) of the current
+frame.
+
+To get the unmasked payload call C<< $frameinfo->unmask($masked_data) >>.
+
 =item $request->in_junk($dir,$data,$eof,$time)
 
 Will be called for legally ignored junk (empty lines) in front of request or 
@@ -1268,7 +1727,7 @@ response body, i.e. which have an implicit and non-changeble content-length of 0
 This constant is an array reference of all response codes which will not have a
 response body, i.e. which have an implicit and non-changeble content-length of 0.
 
-=item parse_hdr_fields($header,\%fields) -> $bad_header
+=item parse_hdrfields($header,\%fields) -> $bad_lines
 
 This function parses the given message header (without request or status line!)
 and extracts the C<key:value> pairs into C<%fields>. Each key in C<%fields> is
@@ -1282,6 +1741,24 @@ It will return any remaining data in C<$header> which could not be interpreted
 as proper C<key:value> pairs. If the message contains no errors it will thus
 return C<''>.
 
+=item parse_reqhdr($string,\%header) -> $bad_header
+
+This will parse the given C<$string> as a request header and extract information
+into \%header. These information then later will be given to
+C<in_request_header>. See there for more details about the contents of the hash.
+
+=item parse_rsphdr($string,\%request,\%header) -> $bad_header
+
+This will parse the given C<$string> as a response header and extract
+information into \%header. These information then later will be given to
+C<in_request_header>. See there for more details about the contents of the
+hash.
+
+C<%request> contains information about the request. One might simple use the
+hash filled by C<parse_reqhdr> here. If not at least the information about
+C<method>, C<expect> and C<upgrade> must be provided because they are needed to
+interpret the response correctly.
+
 =back
 
 =head1 LIMITS
